package bug;


import java.awt.Font;
import java.util.concurrent.CyclicBarrier;

import javax.swing.JFrame;

import visad.ConstantMap;
import visad.DataReference;
import visad.DataReferenceImpl;
import visad.DataRenderer;
import visad.Display;
import visad.FlatField;
import visad.FunctionType;
import visad.GraphicsModeControl;
import visad.Integer1DSet;
import visad.ProjectionControl;
import visad.Real;
import visad.RealType;
import visad.Scalar;
import visad.ScalarMap;
import visad.Text;
import visad.TextControl;
import visad.TextType;
import visad.Tuple;
import visad.TextControl.Justification;
import visad.java3d.DefaultRendererJ3D;
import visad.java3d.DisplayImplJ3D;
import visad.java3d.MouseBehaviorJ3D;
import visad.java3d.ProjectionControlJ3D;

/**
 * My own personal dev version of the bug, but with the addition of lots of 
 * additional switches for experimentation.
 * 
 * Use the build script to reliably run this multiple times and locate the error cases.
 * 
 * Some areas to investigate:
 * 1. Can we detect when the bug is manifest and throw an exception?
 * 	- Is it detectable in the scene graph or in the pixels on the screen?
 * 2. Can we look at the infrastructure and determine where the race/issue is that is causing the bug. 
 * 	- Can we trace the drawing through the infrastructure?
 * 3. Can we reliably reproduce the bug?
 * 	- Can we identify the conditions under which it is more or less likely to occur
 * 	- Can we use statistics to indicate the likelihood of the bug occurring?
 * 
 * Random Observations:
 * - the barrier seems to reduce the occurrence
 * - the zooming fix does not fix it
 * - the geom by ref does not fix it
 * - tested with recent jvm 1.6, j3d 1.5.2, visad 20100628, still occurs
 * 
 */
public class ReproduceVISADBugDev
{
	// time in ms to wait between events in the sequence
	private final static long WAIT_TIME = 500; // ms
	
	// display status of system
	private final static boolean PRINT_STATUS = false;
	// add complexity to the text to change rendering times
	private final static boolean DO_TEXT_PROPERTIES = true;
	// print the scene graph to stdout at the end
	private final static boolean DISPLAY_SCENE_GRAPH = false;
	// display system versions to stdout
	private final static boolean DISPLAY_VERSIONS = false;
	// tests zooming the projection matrix as a fix
	private final static boolean DO_TEST_FIX = false;
	// sets the system property visad.java3d.geometryByRef to true
	private final static boolean DO_SET_GEOM_BY_REF = false;	
	// run the updates in different threads
	private final static boolean DO_IN_DIFFERENT_THREADS = false;
	// use a barrier to wait for the end
	private final static boolean USE_BARRIER = false;

	
	
	private void runBug() throws Exception {
		final Display display = new DisplayImplJ3D("display1");
		// projection
		GraphicsModeControl dispGMC = ((DisplayImplJ3D)display).getGraphicsModeControl();
		dispGMC.setProjectionPolicy(DisplayImplJ3D.PARALLEL_PROJECTION);	
		// frame
		final JFrame jframe = new JFrame("BUGZ!");
		jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		jframe.getContentPane().add(((DisplayImplJ3D)display).getComponent());
		jframe.setSize(640, 480);
		jframe.setVisible(true);
		// barrier
		final CyclicBarrier barrier = new CyclicBarrier(3); // 2 update threads and main
		// line
		final DataReference line = makeLine(display);
		Runnable runLine = new Runnable(){
			public void run(){
				try{addRemoveLine(display, line, barrier);}catch(Exception e){e.printStackTrace();
			}
		}};
		// text
		final DataReference text = makeTextData(display);
		Runnable runText = new Runnable(){
			public void run(){
				try{addRemoveText(display, text, barrier);}catch(Exception e){e.printStackTrace();
			}
		}};
		//
		// 1. Setup and Wait 		
		//
		if (PRINT_STATUS) {
			System.err.println("Visible, waiting to run...");	
		}		
		Thread.sleep(WAIT_TIME);		
		
		//
		// 2. Run		
		//
		if (DO_IN_DIFFERENT_THREADS) {
			// do it (i believe we want them to go at the same time)
			new Thread(runText).start();
			new Thread(runLine).start();				
		} else {
			runText.run();
			Thread.sleep(WAIT_TIME);
			runLine.run();
		}
		if (USE_BARRIER) {
			barrier.await(); // finish
		}
		
		//
		// 3. Done
		//
		if (PRINT_STATUS) {
			System.err.println("Done. The line/text may or may not be visible.");
		}		
		// print scene graph, or whatever.
		if(DISPLAY_SCENE_GRAPH) {
			SceneGraphUtils.printAlternativeSceneGraph(display);
		}
		if(DO_TEST_FIX) {
			runTestZoomToFix(display);
		}
	}

	private void addRemoveText(
			Display display, 
			DataReference dataReference,
			CyclicBarrier barrier) throws Exception{
		// remove
		display.removeReference(dataReference);
		// add
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, dataReference, textColour);
		// wait to finish
		if (USE_BARRIER) {
			barrier.await();
		}
	}

	private void addRemoveLine(
			Display display, 
			DataReference dataReference,
			CyclicBarrier barrier) throws Exception {
		// remove
		display.removeReference(dataReference);
		// add
		int thickness = 4;
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(thickness, Display.LineWidth),new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, dataReference, textColour);
		// wait to finish
		if (USE_BARRIER) {
			barrier.await();
		}
	}
	
	private DataReference makeLine(Display display) throws Exception{
		RealType time = RealType.getRealType("time0");
		RealType height = RealType.getRealType("height0");		
		ScalarMap timeMap = new ScalarMap(time, Display.XAxis);
		ScalarMap heightMap = new ScalarMap(height, Display.YAxis);
		FunctionType func_time_height = new FunctionType(time, height);
		display.addMap(timeMap);
		display.addMap(heightMap);				
		float[][] h_vals = new float[][]{{0.0f, 33.75f, 45.0f, 33.75f, 0.0f,} };	
		Integer1DSet time_set = new Integer1DSet(time, h_vals[0].length);
		FlatField vals_ff1  = new FlatField(func_time_height, time_set);
		vals_ff1.setSamples(h_vals);		
		DataReferenceImpl data = new DataReferenceImpl("data_ref0");
		data.setData(vals_ff1);
		int thickness = 4;
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(thickness, Display.LineWidth),new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, data, textColour);		
		return data;
	}
	
	private DataReference makeTextData(Display display) throws Exception{
		TextType textType = TextType.getTextType("TimeAnnotation0");
		ScalarMap textMap = new ScalarMap(textType, Display.Text);
		display.addMap(textMap);
		TextType trackText = (TextType)textMap.getScalar();		
		float x = 15f; 
		float y = 15f;
		String text = "0000000000000000";
		Scalar[] dataValues = {new Real(RealType.Latitude,x),new Real(RealType.Longitude,y),
				new Text(trackText, text)};		
		Tuple tuple = new Tuple(dataValues);
		DataReferenceImpl data = new DataReferenceImpl("trackLabel0");
		data.setData(tuple);
		ConstantMap[] textColour = {new ConstantMap(1.0, Display.Red),
				new ConstantMap(1.0, Display.Green),new ConstantMap(1.0, Display.Blue),
				new ConstantMap(1.0, Display.ZAxis)};
		DataRenderer dataRenderer = new DefaultRendererJ3D();		
		display.addReferences(dataRenderer, data, textColour);
		if(DO_TEXT_PROPERTIES) {
			addTextProperties(display, textMap);
		}
		return data;
	}
	
	private void addTextProperties(Display display, ScalarMap textMap) throws Exception {
		int size = 6;
		Font newFont = new Font("Serif", Font.PLAIN, size);
		double[] rot = new double[3];
		double[] scale = new double[1];
		double[] trans = new double[3];
		double baseFontSize = 14;		
		ProjectionControlJ3D pcontrol = (ProjectionControlJ3D) ((DisplayImplJ3D) display).getProjectionControl();
		TextControl tcontrol = (TextControl)textMap.getControl();
		MouseBehaviorJ3D.unmake_matrix(rot, scale,trans, pcontrol.getMatrix());			
		tcontrol.setSize((size / baseFontSize) / scale[0]);
		tcontrol.setFont(newFont);
		tcontrol.setRotation(180);
		tcontrol.setJustification(Justification.LEFT);
		tcontrol.setVerticalJustification(Justification.CENTER);		
	}
	
	private void runTestZoomToFix(Display display) throws Exception {
		Thread.sleep(1000);
		ProjectionControl projectionControl = ((DisplayImplJ3D)display).getProjectionControl();
		double[] matrix = projectionControl.getMatrix();		
		double [] base = {matrix[0], matrix[5], matrix[10]}; 
		// progressive zoom
		for (int i = 0; i < 4; i++)
		{
			matrix[0] += 0.25;
			matrix[5] += 0.25;
			matrix[10] += 0.25;
			projectionControl.setMatrix(matrix);
		}		
		// reset
		matrix[0] = base[0];
		matrix[5] = base[1];
		matrix[10] = base[2];
		projectionControl.setMatrix(matrix);		
	}

	public static void main(String[] args)
	{
		try{
			// suggested by rink@ssec.wisc.edu
			if (DO_SET_GEOM_BY_REF) {
				System.setProperty("visad.java3d.geometryByRef", "true");				
				System.err.println("visad.java3d.geometryByRef = " + System.getProperty("visad.java3d.geometryByRef"));
			}
			// display version information
			if(DISPLAY_VERSIONS) {
				Versions.versions();
			}
			// run the bug
			new ReproduceVISADBugDev().runBug();
		}catch(Exception e)
		{
			System.err.println("App Exception");
			e.printStackTrace();
		}
	}
}