package bug;

import java.util.Enumeration;

import javax.media.j3d.BoundingBox;
import javax.media.j3d.Geometry;
import javax.media.j3d.Group;
import javax.media.j3d.LineArray;
import javax.media.j3d.Node;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

import visad.Display;
import visad.DisplayImpl;
import visad.DisplayRenderer;
import visad.java3d.DisplayRendererJ3D;

/**
 * Print the scene graph Based on :
 * http://www.mail-archive.com/java3d-interest@java.sun.com/msg00146.html
 * <br />
 * Convert a Java3D scene graph into a string. Expects a VisAD display (DisplayImpl) that has a 
 * DisplayRendererJ3D renderer.
 * <br />
 * Adapted and extended for ease of use, see: alternativeSceneGraphToString()
 * 
 * @author brownlee
 * 
 */
public class SceneGraphUtils
{
	
	/**
	 * Convert a Java3D scene graph into a string.
	 * 
	 * @param display
	 * @return
	 */
	public static String sceneGraphToString(Display display)
	{		
		DisplayRenderer renderer = ((DisplayImpl)display).getDisplayRenderer();
		if(!(renderer instanceof DisplayRendererJ3D)) {
			throw new RuntimeException("Renderer is not the expected type: " + DisplayRendererJ3D.class.getName());
		}		
		Node root = ((DisplayRendererJ3D)renderer).getRoot();
		return sceneGraphToString(root);
	}
	
	/**
	 * 
	 * @param display
	 */
	public static void printSceneGraph(Display display) {
		System.out.println(sceneGraphToString(display));
	}
	
	/**
	 * Convert a Java3D scene graph into a string.
	 * 
	 * @param n
	 * @return
	 */
	public static String sceneGraphToString(Node n)
	{		
		StringBuilder out = new StringBuilder();
		out.append("Start\n");
		dumpNodes(n, 0, out);
		return out.toString();
	}

	
	private static void dumpTG(TransformGroup tg, StringBuilder out)
	{
		Transform3D t = new Transform3D();
		Vector3d v = new Vector3d();
		tg.getTransform(t);
		t.get(v);
		out.append(" x=" + v.x + " y=" + v.y + " z=" + v.z);
	}

	private static void dumpBB(BoundingBox bb, StringBuilder out)
	{
		Point3d p = new Point3d();
		bb.getLower(p);
		out.append(" l=" + p);
		bb.getUpper(p);
		out.append(" u=" + p);
	}
	private static void dumpShape(Shape3D shape, StringBuilder out)
	{
		Geometry geo = shape.getGeometry();
		out.append(" G=" + shape.getGeometry().getClass().getName());
	}
	

	private static void dumpNodes(Node n, int depth, StringBuilder out)
	{
		int i;
		for (i = 0; i < depth; i++){
			out.append('.');
		}	

		// dump
		out.append(n.getClass().getName());
		if (n instanceof TransformGroup){
			dumpTG((TransformGroup) n, out);
		} else if (n instanceof Shape3D) {
			dumpShape((Shape3D)n, out);
		}			
		out.append("\n");

		// process children
		if (n instanceof Group)
		{
			Enumeration children = ((Group) n).getAllChildren();

			while (children.hasMoreElements())
			{
				Object child = children.nextElement();
				if (child instanceof Node)
					dumpNodes((Node) child, depth + 1, out);
			}
		}
	}
	
	
	
	public static void printAlternativeSceneGraph(Display display) {
		System.out.println(alternativeSceneGraphToString(display));
	}
	
	/**
	 * An alternative, custom version that strips the scene graph down to its
	 * interesting geometry.
	 * 
	 * @param display
	 * @return
	 */
	public static String alternativeSceneGraphToString(Display display){
		DisplayRenderer renderer = ((DisplayImpl)display).getDisplayRenderer();
		if(!(renderer instanceof DisplayRendererJ3D)) {
			throw new RuntimeException("Renderer is not the expected type: " + DisplayRendererJ3D.class.getName());
		}		
		Node root = ((DisplayRendererJ3D)renderer).getRoot();
		StringBuffer buf = new StringBuffer();
		buf.append("START\n");
		collectSceneGraph(buf, root, 0);
		return buf.toString();
	}
	
	private final static void collectSceneGraph(StringBuffer buf, Node node, int lvl)
	{
		if (node instanceof Group) {
			Enumeration children = ((Group)node).getAllChildren();
			while (children.hasMoreElements()) {
				collectSceneGraph(buf, (Node) children.nextElement(), lvl+1);
			}
		} else {			
			for (int i = 0; i < lvl; i++) {buf.append("-");}	
			buf.append(node.getClass().getName() + " ");	
			
			if(node instanceof Shape3D) {
				Shape3D shape = (Shape3D) node;
				Geometry geo = shape.getGeometry();				
				// shape data 
				buf.append("Live="+shape.isLive()+", ");
				buf.append("Geometry="+geo+", ");
				buf.append("GeometryLive="+geo.isLive()+", ");
				if(geo instanceof LineArray) {
					LineArray la = (LineArray) geo;
					// TODO
				}
			} else {
				// skip				
			}	
			
			buf.append("\n");
		}
	}
}